#!/usr/bin/python
from pyhydrophone.hydrophone import Hydrophone

import os
import numpy as np
import soundfile as sf
import datetime

try:
    import scipy.signal as sig
except ModuleNotFoundError:
    pass


class BruelKjaer(Hydrophone):
    """
    Init an instance of B&K Nexus.
    Check well the Vpp in case you don't have a reference signal! Specially of the recorder used.

    Parameters
    ----------
    name: str
        Name of the acoustic recorder
    model: str or int
        Model of the acoustic recorder
    serial_number : str or int
        Serial number of the acoustic recorder
    preamp_gain: float
        Amplification selected in the Nexus in db 10*log10((V/uPa)^2)
    Vpp: float
        Volts peak to peak
    string_format: string
        Format of the datetime string present in the filename
    type_signal : str
        Can be 'ref' or 'test'
    calibration_file : string or Path
        File where the frequency dependent sensitivity values for the calibration are
    """
    def __init__(self, name, model, serial_number, preamp_gain, Vpp=2.0, string_format="%y%m%d%H%M%S", type_signal='ref',
                 max_calibration_time=120.0, calibration_file=None, **kwargs):
        self.amplif = np.sqrt(10**(preamp_gain/10)) * 1e6
        self.preamp_gain = preamp_gain
        self.type_signal = type_signal
        self.max_calibration_time = max_calibration_time
        self.cal_freq = 159.9
        self.min_duration = 10.0
        
        super().__init__(name, model, serial_number, sensitivity=0.0, preamp_gain=preamp_gain,
                         Vpp=Vpp, string_format=string_format, calibration_file=calibration_file, **kwargs)

    def __setattr__(self, name, value):
        """
        If the amplif is changed, update the preamp_gain
        """
        if name == 'amplif':
            self.__dict__['preamp_gain'] = 10 * np.log10((value/1e6)**2)
        if name == 'preamp_gain':
            self.__dict__['amplif'] = np.sqrt(10**(value/10)) * 1e6
        if name == 'type_signal':
            if value not in ['ref', 'test']:
                raise Exception('%s is not an option in Nexus!' % value)
        return super().__setattr__(name, value)

    def get_name_datetime(self, file_name):
        """
        Get the data and time of recording from the name of the file

        Parameters
        ----------
        file_name : string
            File name (not path) of the file
        """
        name = file_name.split('.')
        date_string = name[0].split('_')[0]
        date = super().get_name_datetime(date_string)
        return date

    def get_new_name(self, filename, new_date):
        """
        Replace the datetime with the appropriate one

        Parameters
        ----------
        filename : string
            File name (not path) of the file
        new_date : datetime object
            New datetime to be replaced in the filename
        """
        old_date = self.get_name_datetime(filename)
        old_date_name = datetime.datetime.strftime(old_date, self.string_format)
        new_date_name = datetime.datetime.strftime(new_date, self.string_format)
        new_filename = filename.replace(old_date_name, new_date_name)
        
        return new_filename

    def read_start_time_metadata(self, file_path):
        """
        Return the starting time of the file by getting the last modification minus the duration of the file

        Parameters
        ----------
        file_path : string or Path
            Path to the file to read the information from

        Returns
        -------
        Datetime, starting moment of the file
        """
        modif_timestamp = os.path.getmtime(file_path)
        modif_datetime = datetime.datetime.fromtimestamp(modif_timestamp)
        duration = datetime.timedelta(seconds=sf.info(file_path).duration)
        start_time = modif_datetime - duration
        return start_time
    
    def update_calibration(self, ref_signal):
        """
        Update the calibration

        Parameters
        ----------
        ref_signal : str or Path
            File path to the reference file to update the Vpp according to the calibration tone
        """
        ref_wav = np.sqrt((ref_signal ** 2).mean())
        # Sensitivity is in negative values!
        if self.type_signal == 'ref':
            # The signal is then 1 V
            ref_v = 1
        else:
            # The signal is the amplification value
            ref_v = 100 * self.amplif * np.sqrt(2)
        self.preamp_gain = 10 * np.log10((self.amplif / 1e6) ** 2) + 10*np.log10((ref_wav / ref_v)**2)

    def calibrate(self, file_path):
        """
        Find the beginning and ending sample of the calibration tone
        Returns start and end points, in seconds

        Parameters
        ----------
        file_path : string or Path
            File where to look for the calibration (at the beginning of the file)

        Returns
        -------
        end sample of the calibration (int)
        """
        wav_file = sf.SoundFile(file_path)
        tone_samples = int(self.max_calibration_time * wav_file.samplerate)
        first_part = wav_file.read(frames=tone_samples)
        analytic_signal = sig.hilbert(first_part)
        amplitude_envelope = np.abs(analytic_signal)
        possible_points = np.zeros(amplitude_envelope.shape)
        possible_points[np.where(amplitude_envelope >= 0.05)] = 1
        start_points = np.where(np.diff(possible_points) == 1)[0]
        end_points = np.where(np.diff(possible_points) == -1)[0]
        if start_points.size == 0:
            return None
        if end_points[0] < start_points[0]:
            end_points = end_points[1:]
        if start_points.size != end_points.size:
            start_points = start_points[0:end_points.size]
        select_idx = np.argmax(end_points - start_points)
        # Round to a second
        start = int(start_points[select_idx])
        end = int(end_points[select_idx])

        if (end - start) / wav_file.samplerate < self.min_duration:
            return 0

        self.update_calibration(first_part[start:end])

        return end



#!/usr/bin/python
from datetime import datetime
import numpy as np
import soundfile as sf
import pandas as pd
import scipy.interpolate

try:
    import scipy.signal as sig
except ModuleNotFoundError:
    pass


class Hydrophone:
    """
    Base class Hydrophone initialization

    Parameters
    ----------
    name: str
        Name of the acoustic recorder
    model: str or int
        Model of the acoustic recorder
    serial_number : str or int
        Serial number of the acoustic recorder
    sensitivity : float
        Sensitivity of the acoustic recorder in db
    preamp_gain : float
        Gain of the preamplifier in dB
    Vpp : float
        Voltage peak to peak in volts
    string_format : string
        Format of the datetime string present in the filename
    calibration_file : string or Path
        File where the frequency dependent sensitivity values for the calibration are
    """

    def __init__(self, name, model, serial_number, sensitivity, preamp_gain, Vpp, string_format, calibration_file=None,
                 **kwargs):
        self.name = name
        self.model = model
        self.serial_number = serial_number
        try:
            self.sensitivity = float(sensitivity)
        except ValueError:
            raise ValueError("Sensitivity must be an integer or float")
        try:
            self.preamp_gain = float(preamp_gain)
        except ValueError:
            raise ValueError("preamp_gain must be an integer or float")
        try:
            self.Vpp = float(Vpp)
        except ValueError:
            raise ValueError("Vpp must be an integer or float")
        self.string_format = string_format
        self.cal_freq = 250
        self.cal_value = 114
        self.calibration_file = calibration_file
        self.freq_cal = None
        if calibration_file is not None:
            self.get_freq_cal(**kwargs)

    def get_name_datetime(self, date_string):
        """
        Read the name of the file and according to the hydrophone protocol get the date

        Parameters
        ----------
        date_string : string
            Datetime in string format
        """
        date = datetime.strptime(date_string, self.string_format)
        return date

    def get_new_name(self, filename, new_date):
        """
        Replace the datetime with the appropriate one

        Parameters
        ----------
        filename : string
            File name (not path) of the file
        new_date : datetime object
            New datetime to be replaced in the filename
        """
        raise Exception('get_new_name should be declared in the subclass %s' % self.__name__)

    def change_calibration_system(self, cal_freq, cal_val):
        """
        Change the parameters of the calibration system (piston phone)

        Parameters
        ----------
        cal_freq: float
            Calibration frequency in Hz
        cal_val: float
            Expected value in db

        """
        self.cal_value = cal_val
        self.cal_freq = cal_freq

    def update_calibration(self, calibration_signal, p_ref=1.0):
        """
        Updates ONLY the parameter preamp_gain of the hydrophone with a correction factor to match expected calibration.

        Parameters
        ----------
        calibration_signal: np.array
            signal to calibrate from (already cut to ONLY calibration)
        p_ref: float
            Reference pressure to compute db from

        Returns
        -------
        Updates the parameter preamp_gain
        """
        ref_wav = np.sqrt((calibration_signal ** 2).mean())
        mv = 10 ** (self.sensitivity / 20.0) * p_ref
        ma = 10 ** (self.preamp_gain / 20.0) * p_ref
        gain_upa = (self.Vpp / 2.0) / (mv * ma)
        real_db = 20 * np.log10(ref_wav * gain_upa)
        correction_factor = real_db - self.cal_value
        self.preamp_gain += correction_factor

    def calibrate(self, file_path):
        """
        Find the beginning and ending sample of the calibration tone
        Returns start and end points, in seconds

        Parameters
        ----------
        file_path : string or Path
            File where to look for the calibration (at the beginning of the file)

        Returns
        -------
        end sample of the calibration (int)
        """
        wav_file = sf.SoundFile(file_path)
        calibration_signal = wav_file.read()
        self.update_calibration(calibration_signal)

        return None

    def end_to_end_calibration(self, p_ref=1.0):
        """
        Returns the end to end calibration of the system, so it can be directly used on a wav file to obtain uPa

        Parameters
        ----------
        p_ref

        Returns
        -------
        End to end calibration in db
        """
        mv = 10 ** (self.sensitivity / 20.0) * p_ref
        ma = 10 ** (self.preamp_gain / 20.0) * p_ref
        gain_upa = (self.Vpp / 2.0) / (mv * ma)
        return 10 * np.log10(gain_upa ** 2)

    def get_freq_cal(self, val='sensitivity', sep=',', freq_col_id=0, val_col_id=1, start_data_id=0):
        """
        Compute a dataframe with all the frequency dependent sensitivity values from the calibration file

        Parameters
        ----------
        val : str
            Can be 'sensitivity' or 'end_to_end' depending on what are the values in the calibration file
        sep : str
            Separator between the different columns in csv or txt files
        freq_col_id : int
            Id of the frequency column in the file (starts with 0)
        val_col_id : int
            Id of the values column in the file (starts with 0)
        start_data_id : int
            Id of the first line with data (without title) in the file (starts with 0)
        """

        if self.calibration_file.suffix == '.csv' or self.calibration_file.suffix == '.txt':
            df = pd.read_csv(self.calibration_file, sep=sep, header=None)

        elif self.calibration_file.suffix == '.xlsx':
            df = pd.read_excel(self.calibration_file, header=None)

        df = df.iloc[:, (i for i in range(len(df.columns)) if i == freq_col_id or i == val_col_id)]
        df = df[start_data_id:]
        df = df.dropna(subset=[df.columns[0]])
        df = df.replace('[A-Za-z:]', '', regex=True).astype(float)
        df = df.reset_index(drop=True)
        df.columns = ['frequency', val]

        self.freq_cal = df

    def freq_cal_inc(self, frequencies, p_ref=1.0):
        """
        Returns a dataframe with the frequency dependent values to increment from the selected frequencies you give from
        the data you want to increment

        Parameters
        ----------
        frequencies : 1d array
            Frequencies from the data you want to increment with frequency dependent calibration
        p_ref: float
            Reference pressure to compute db from

        Returns
        -------
        df_freq_inc : pandas Dataframe
            Frequency dependent values to increment in your data
        """
        df = self.freq_cal
        val = df.columns[1]
        min_freq = df['frequency'][0]
        max_freq = df['frequency'][df.shape[0] - 1]
        interpol = scipy.interpolate.interp1d(df['frequency'], df[val], kind='linear')

        frequencies_below = frequencies.compress(frequencies < min_freq)
        frequencies_between = frequencies.compress(np.logical_and(frequencies >= min_freq, frequencies <= max_freq))
        frequencies_above = frequencies.compress(frequencies > max_freq)

        freq_dep_cal = interpol(frequencies_between)

        if val == 'sensitivity':
            mv = 10 ** (freq_dep_cal / 20.0) * p_ref
            ma = 10 ** (self.preamp_gain / 20.0) * p_ref
            gain_upa = (self.Vpp / 2.0) / (mv * ma)
            freq_cal_inc = 10 * np.log10(gain_upa ** 2) - self.end_to_end_calibration()
        elif val == 'end_to_end':
            freq_cal_inc = freq_dep_cal - self.end_to_end_calibration()
        else:
            raise ValueError(f'columns name {val} is not implemented. Only end_to_end or sensitivity are valid values')
        freq_cal_inc = np.concatenate(
            (np.zeros(frequencies_below.shape), freq_cal_inc, np.zeros(frequencies_above.shape)))
        df_freq_inc = pd.DataFrame(data=np.vstack((frequencies, freq_cal_inc)).T, columns=['frequency', 'inc_value'])

        return df_freq_inc



#!/usr/bin/python
from pyhydrophone.hydrophone import Hydrophone

import os
import zipfile
import numpy as np
import pandas as pd
import soundfile as sf
from datetime import datetime
import xml.etree.ElementTree as ET
import xml
import requests
import pathlib


class SoundTrap(Hydrophone):
    """
    Initialize a SoundTrap instance

    Parameters
    ----------
    name: str
        Name of the acoustic recorder
    model: str or int
        Model of the acoustic recorder
    serial_number : str or int
        Serial number of the acoustic recorder. It has to match the one in the calibration file
    sensitivity : float
        Sensitivity of the acoustic recorder in db. If None the one from the calibration file will be read
    Vpp : float
        Value will be ignored and always 2 will be used. Kept for compatibility with other instruments in pipelines
    gain_type : str
        'High' or 'Low', depending on the settings of the recorder
    string_format : string
        Format of the datetime string present in the filename
    calibration_file : string or Path
        File where the frequency dependent sensitivity values for the calibration are
    """
    def __init__(self, name, model, serial_number, sensitivity=None, Vpp=2, gain_type='High',
                 string_format="%y%m%d%H%M%S", calibration_file=None,  **kwargs):
        if sensitivity is None:
            try:
                query = 'http://oceaninstruments.azurewebsites.net/api/Devices/Search/%s' % serial_number
                response = requests.get(query).json()
                if len(response) > 1:
                    models_available = {}
                    for device in response:
                        if device['serialNo'] == str(serial_number):
                            models_available[device['modelName']] = device['deviceId']

                    if model not in models_available.keys():
                        raise AttributeError('There are multiple instruments with serial number %s. Set the model '
                                             'parameter to match the model specified in the SoundTrap calibration '
                                             'webpage to chose the correct one: %s' %
                                             (serial_number, models_available.keys()))
                    else:
                        device_id = models_available[model]
                else:
                    # Ignore the model name if there is only one serial number
                    device_id = response[0]['deviceId']
                query = 'http://oceaninstruments.azurewebsites.net/api/Calibrations/Device/%s' % device_id
                response = requests.get(query).json()[0]
                if gain_type == 'High':
                    sensitivity = -response['highFreq']
                elif gain_type == 'Low':
                    sensitivity = -response['lowFreq']
                else:
                    raise Exception('Gain type %s is not implemented!' % gain_type)
            except ConnectionError:
                raise Exception('Serial number %s is not in the OceanInstruments database!' % serial_number)

        super().__init__(name, model, serial_number=serial_number, sensitivity=sensitivity, preamp_gain=0.0,
                         Vpp=2.0, string_format=string_format, calibration_file=calibration_file, **kwargs)

    @staticmethod
    def read_file_specs(xmlfile_path, last_gain, date_format='%Y-%m-%dT%H:%M:%S'):
        """
        Read the specs of the recording from the XML file and save them to the object

        Parameters
        ----------
        xmlfile_path : string or path
            Path to the xml file
        last_gain : str
            Last gain type. 'High' or 'Low', depending on the settings of the recorder
        date_format : string
            Format of the datetime in the .log.xml file
        """
        tree = ET.parse(xmlfile_path)
        type_start = tree.find('EVENT/START').get('STATE')

        # Metadata colected
        temp = float(tree.find('EVENT/TEMPERATURE').text)/100

        # WavFileHandler information
        sampling_attr = {}
        WavFileHandler = tree.findall('PROC_EVENT/WavFileHandler')
        for wfh in WavFileHandler:
            sampling_attr.update(wfh.attrib)

        # Info about the sampling
        fs = float(tree.find('CFG/FS').text)

        # Setup information. Read SoundTrap gain ('HIGH' or 'LOW')
        if type_start == 'NEW':
            st_gain = tree.find('EVENT/AUDIO').get('Gain')
        else:
            if last_gain is None:
                print('Unknown gain if it is reopened and the last gain is not passed!')
            st_gain = last_gain

        start_time = datetime.strptime(sampling_attr['SamplingStartTimeLocal'], date_format)
        stop_time = datetime.strptime(sampling_attr['SamplingStopTimeLocal'], date_format)

        return {'type_start': type_start, 'temp': temp, 'fs': fs, 'st_gain': st_gain,
                'start_time': start_time, 'stop_time': stop_time}

    def get_name_datetime(self, file_name):
        """
        Get the data and time of recording from the name of the file
        Will convert the local in UTC. It assumes the localtime is the one from the computer

        Parameters
        ----------
        file_name : string
            File name (not path) of the file
        """
        name = file_name.split('.')
        date_string = name[1]
        date = super().get_name_datetime(date_string)
        return date

    @staticmethod
    def get_xml_utc_datetime(file_path):
        """
        Get the UTC datetime from the xml file

        Parameters
        ----------
        file_path : str or Path

        """
        if type(file_path) == str:
            xml_name = file_path.replace('.wav', '.log.xml')
        else:
            xml_name = file_path.parent.joinpath(file_path.name.replace('.wav', '.log.xml'))

        tree = ET.parse(xml_name)
        WavFileHandler_list = tree.findall('PROC_EVENT/WavFileHandler')
        for wfh in WavFileHandler_list:
            if 'SamplingStartTimeUTC' in wfh.attrib.keys():
                utc_datetime = datetime.strptime(wfh.attrib.values(), format='%Y-%m-%dTH:M:S')
                return utc_datetime

        return None

    def get_new_name(self, filename, new_date):
        """
        Replace the datetime with the appropriate one

        Parameters
        ----------
        filename : string
            File name (not path) of the file
        new_date : datetime object
            New datetime to be replaced in the filename
        """
        old_date = self.get_name_datetime(filename)
        old_date_name = datetime.strftime(old_date, "%y%m%d%H%M%S")
        new_date_name = datetime.strftime(new_date, "%y%m%d%H%M%S")
        new_filename = filename.replace(old_date_name, new_date_name)

        return new_filename

    def test_calibration(self, signal):
        """
        Test the calibration of the soundtrap
        """
        # TO BE IMPLEMENTED


class SoundTrapHF(SoundTrap):
    def __init__(self, name, model, serial_number, sensitivity=None, gain_type='High', Vpp=2,
                 string_format="%y%m%d%H%M%S", calibration_file=None, **kwargs):
        """
        Init a SoundTrap HF reader

        Parameters
        ----------
        name: str
            Name of the acoustic recorder
        model: str or int
            Model of the acoustic recorder
        serial_number : str or int
            Serial number of the acoustic recorder. It has to match the one in the calibration file
        sensitivity : float
            Sensitivity of the acoustic recorder in db. If None the one from the calibration file will be read
        gain_type : str
            'High' or 'Low', depending on the settings of the recorder
        Vpp : float
            Value will be ignored and always 2 will be used. Kept for compatibility with other instruments in pipelines
        string_format : string
            Format of the datetime string present in the filename
        calibration_file : string or Path
            File where the frequency dependent sensitivity values for the calibration are
        """
        super().__init__(name=name, model=model, serial_number=serial_number, sensitivity=sensitivity,
                         gain_type=gain_type, Vpp=2, string_format=string_format,
                         calibration_file=calibration_file, **kwargs)

    def read_HFfolder(self, main_folder_path, zip_mode=False, include_dirs=False):
        """
        Read all the clicks in all the folders. If zip_mode is True and include_dirs is True, only the INSIDE folders
        can be zipped inside a non-zipped folder. If only one zip folder is to be analyzed, then set include_dirs
        to False.

        Parameters
        ----------
        main_folder_path: str or Path
            Folder containing all the files and/or subfolders to be extracted
        zip_mode : boole
            Set to True if the folders are zipped
        include_dirs : bool
            Set to True if folder needs to be analyzed recursively

        Returns
        -------
        A DataFrame with all the clicks of all the folders and a fs metadata parameter with the sampling rate
        """
        if type(main_folder_path) == str:
            main_folder_path = pathlib.Path(main_folder_path)
        clicks = pd.DataFrame()
        if not zip_mode:
            if include_dirs:
                glob_str = '**.wav'
            else:
                glob_str = '*.wav'
            for file_name in main_folder_path.glob(glob_str):
                clicks_file = self.read_HFclicks_file(file_name, zip_mode)
                clicks = pd.concat([clicks, clicks_file])
        else:
            if include_dirs:
                for zipped_dir in main_folder_path.glob('*'):
                    clicks_file = self.read_HFfolder(zipped_dir, zip_mode=zip_mode, include_dirs=False)
                    clicks = clicks.append(clicks_file)
            else:
                folder_path = zipfile.ZipFile(main_folder_path, 'r', allowZip64=True)
                files_list = folder_path.namelist()
                for file_name in files_list:
                    if file_name.split('.')[-1] == 'wav':
                        file_name = pathlib.Path(folder_path.filename).joinpath(file_name)
                        clicks_file = self.read_HFclicks_file(file_name, zip_mode)
                        clicks = clicks.append(clicks_file)

        return clicks

    def read_HFclicks_file(self, wavfile_path, zip_mode=False):
        """
        Read all the clicks stored in a folder with soundtrap files

        Parameters
        ----------
        wavfile_path: str
            Wav file path
        zip_mode : boolean
            Set to True if the folders are zipped

        Returns
        -------
        A DataFrame with all the clicks and a fs metadata parameter with the sampling rate
        """
        if not isinstance(wavfile_path, pathlib.Path):
            wavfile_path = pathlib.Path(wavfile_path)
        bcl_name = wavfile_path.name.replace('.wav', '.bcl')
        dwv_name = wavfile_path.name.replace('.wav', '.dwv')
        xml_name = wavfile_path.name.replace('.wav', '.log.xml')
        if zip_mode:
            zip_file = zipfile.ZipFile(wavfile_path.parent, 'r', allowZip64=True)
            bcl_path = zip_file.open(bcl_name)
            dwv_path = zip_file.open(dwv_name)
            xml_path = zip_file.open(xml_name)
        else:
            bcl_path = os.path.join(wavfile_path.parent, bcl_name)
            dwv_path = os.path.join(wavfile_path.parent, dwv_name)
            xml_path = os.path.join(wavfile_path.parent, xml_name)

        try:
            file_clicks = self._read_HFclicks(bcl_path, dwv_path, xml_path)
        except FileNotFoundError:
            print(dwv_path, 'has some problem and can not be read')
        if zip_mode:
            dwv_path = wavfile_path.parent.joinpath(bcl_name)
        file_clicks['filename'] = str(dwv_path)
        return file_clicks

    def _read_HFclicks(self, bcl_path, dwv_path, xml_path):
        """
        Read the clicks of one soundtrap file

        Parameters
        ----------
        bcl_path : str or Path
            Path to the bcl file
        dwv_path : str or Path
            Path to the dwv file
        xml_path : str or Path
            Path to the .log.xml file

        Returns
        -------
        A DataFrame with all the parameters from the bcl file + a column with the wave and a column with the datetime
        """

        # Read the wav file with all the clicks
        sound_file = sf.SoundFile(dwv_path, 'r')

        # click_len has to be checked automatically
        try:
            click_len = self.read_HFparams(xml_path=xml_path)
            # Read the info of clicks
            clicks_info = pd.read_csv(bcl_path)
            clicks_info = clicks_info[clicks_info['report'] == 'D']
            clicks_info = clicks_info[clicks_info['state'] == 1]
            waves = []

            for block in sound_file.blocks(blocksize=click_len):
                waves.append(block.astype(float))

            print(dwv_path, 'bcl:', len(clicks_info), 'dwv:', len(waves))

            if len(waves) < len(clicks_info):
                # Cut the clicks info if there are not enough snippets
                clicks_info = clicks_info.loc[0:len(waves)]

            clicks_info['wave'] = waves[0:len(clicks_info)]
            clicks_info['start_sample'] = np.arange(len(clicks_info)) * click_len
            clicks_info['end_sample'] = clicks_info['start_sample'] + click_len
            clicks_info['duration'] = click_len
            clicks_info['fs'] = sound_file.samplerate
            clicks_info['datetime'] = pd.to_datetime(clicks_info['rtime'] + clicks_info['mticks'] / 1e6, unit='s')

            return clicks_info.reset_index(drop=True)
        except xml.etree.ElementTree.ParseError:
            print(f'XML file {xml_path} could not be parsed, ignoring...')
            return pd.DataFrame()


    @staticmethod
    def read_HFparams(xml_path):
        """
        Return the length of the clips and the time in between

        Parameters
        ----------
        xml_path : string or Path
            Path to the .log.xml file

        Returns
        -------
        Clip length in samples (int)
        """
        tree = ET.parse(xml_path)

        # blank_len = int(tree.find('CFG/BLANKING').text)
        clip_len = int(tree.find('CFG/PREDET').text) + int(tree.find('CFG/POSTDET').text)

        return clip_len



#!/usr/bin/python
from pyhydrophone.hydrophone import Hydrophone

from datetime import datetime

try:
    import matplotlib.pyplot as plt
    import pandas as pd
except ModuleNotFoundError:
    pass


class uAural(Hydrophone):
    """
    Init an instance of uAural

    Parameters
    ----------
    name: str
        Name of the acoustic recorder
    model: str or int
        Model of the acoustic recorder
    serial_number : str or int
        Serial number of the acoustic recorder
    sensitivity : float
        Sensitivity of the acoustic recorder in db
    preamp_gain : float
        Gain of the preamplifier in dB
    Vpp : float
        Voltage peak to peak in volts
    string_format : string
        Format of the datetime string present in the filename
    calibration_file : string or Path
        File where the frequency dependent sensitivity values for the calibration are
    """
    def __init__(self, name, model, serial_number, sensitivity, preamp_gain, Vpp, string_format="%H%M%S_%Y%m%d",
                 calibration_file=None, **kwargs):
        super().__init__(name, model, serial_number, sensitivity, preamp_gain, Vpp, string_format, calibration_file,
                         **kwargs)

    def get_name_datetime(self, file_name):
        """
        Get the data and time of recording from the name of the file

        Parameters
        ----------
        file_name : string
            File name (not path) of the file
        """
        name = file_name.split('.')[0]
        date_and_time = name.split('_')[:2]
        date_string = '_'.join(date_and_time)
        date = super().get_name_datetime(date_string)
        return date

    def get_new_name(self, filename, new_date):
        """
        Replace the datetime with the appropriate one

        Parameters
        ----------
        filename : string
            File name (not path) of the file
        new_date : datetime object
            New datetime to be replaced in the filename
        """
        old_date = self.get_name_datetime(filename)
        old_date_name = datetime.strftime(old_date, self.string_format)
        new_date_name = datetime.strftime(new_date, self.string_format)
        new_filename = filename.replace(old_date_name, new_date_name)

        return new_filename



from pyhydrophone.upam import uPam
from pyhydrophone.amar import AmarG3, AmarG3MEMS
from pyhydrophone.soundtrap import SoundTrap, SoundTrapHF
from pyhydrophone.bruelkjaer import BruelKjaer
from pyhydrophone.mte import MTE
from pyhydrophone.rtsys import RTSys
from pyhydrophone.ears import EARs
from pyhydrophone.icListen import icListen
from pyhydrophone.uaural import uAural



#!/usr/bin/python
from pyhydrophone.hydrophone import Hydrophone

from datetime import datetime
import struct
import numpy as np
import zipfile
import os
import pathlib

try:
    import matplotlib.pyplot as plt
    import pandas as pd
except ModuleNotFoundError:
    pass


class RTSys(Hydrophone):
    """
    Init an instance of RTSys

    Parameters
    ----------
    name: str
        Name of the acoustic recorder
    model: str or int
        Model of the acoustic recorder
    serial_number : str or int
        Serial number of the acoustic recorder
    sensitivity : float
        Sensitivity of the acoustic recorder in db
    preamp_gain : float
        Gain of the preamplifier in dB
    Vpp : float
        Voltage peak to peak in volts
    mode: string
        Can be 'lowpower' or 'broadband'
    channel: string
        Channel to process, 'A', 'B', 'C' or 'D'
    string_format : string
        Format of the datetime string present in the filename
    calibration_file : string or Path
        File where the frequency dependent sensitivity values for the calibration are
    """

    def __init__(self, name, model, serial_number, sensitivity, preamp_gain, Vpp, mode, channel='A',
                 string_format="%Y-%m-%d_%H-%M-%S", calibration_file=None):
        super().__init__(name, model, serial_number, sensitivity, preamp_gain, Vpp, string_format, calibration_file,
                         val='sensitivity', sep=';', freq_col_id=0, val_col_id=1, start_data_id=0)
        self.cal_freq = 250
        self.cal_value = 114
        self.mode = mode
        self.channel = channel

    def get_name_datetime(self, file_name):
        """
        Get the data and time of recording from the name of the file

        Parameters
        ----------
        file_name : string
            File name (not path) of the file
        """
        name = file_name.split('.')[0]
        start_timestamp = name.find('_') + 1
        date_string = name[start_timestamp::]
        date = super().get_name_datetime(date_string)
        return date

    def get_new_name(self, filename, new_date):
        """
        Replace the datetime with the appropriate one

        Parameters
        ----------
        filename : string
            File name (not path) of the file
        new_date : datetime object
            New datetime to be replaced in the filename
        """
        old_date = self.get_name_datetime(filename)
        old_date_name = datetime.strftime(old_date, self.string_format)
        new_date_name = datetime.strftime(new_date, self.string_format)
        new_filename = filename.replace(old_date_name, new_date_name)

        return new_filename

    @staticmethod
    def _parse_board_file(board_file_path):
        board_info = pd.read_csv(board_file_path, delimiter=';', names=['id', 'T', 'V', 'I', 'P'],
                                 usecols=[0, 1, 2, 3, 4])
        board_info['T'] = board_info['T'].str.replace('T:', '')
        board_info.loc[board_info['T'] == ''] = np.nan
        board_info['T'] = board_info['T'].astype(float)
        board_info['V'] = board_info['V'].str.replace('V:', '')
        board_info.loc[board_info['V'] == ''] = np.nan
        board_info['V'] = board_info['V'].astype(float)
        board_info['I'] = board_info['I'].str.replace('I:', '')
        board_info.loc[board_info['I'] == ''] = np.nan
        board_info['I'] = board_info['I'].astype(float)
        board_info['P'] = board_info['P'].str.replace('P:', '')
        board_info.loc[board_info['P'] == ''] = np.nan
        board_info['P'] = board_info['P'].astype(float)
        board_info['timestamp'] = pd.to_datetime(board_info['id'].str.replace('@:', '').astype(float), unit='s')
        board_info['dP'] = board_info['timestamp'].diff().dt.total_seconds() * board_info['P']

        return board_info

    def plot_consumption(self, board_file_path):
        """
        Plot the consumption evolution from the board_file_path

        Parameters
        ----------
        board_file_path : str or Path
        """
        board_info = self._parse_board_file(board_file_path)
        board_info.plot(y=['V', 'P'])
        plt.show()

    def plot_consumption_total_mission(self, mission_folder_path, ax=None, show=True):
        if not isinstance(mission_folder_path, pathlib.Path):
            mission_folder_path = pathlib.Path(mission_folder_path)

        total_board = pd.DataFrame()
        for board_file_i in mission_folder_path.glob('**/*.txt'):
            if 'board' in board_file_i.name:
                board_i = self._parse_board_file(board_file_i)
                total_board = pd.concat([total_board, board_i])

        if ax is None:
            fig, ax = plt.subplots()
        total_board.plot(x='timestamp', y=['V', 'P'], secondary_y='P', ax=ax)
        if show:
            plt.show()

    def compute_consumption(self, board_file_path):
        """
        Calculate the total energy consumption of the file

        Parameters
        ----------
        board_file_path : str or Path

        Returns
        -------
        Total consumption in the file
        """
        board_info = self._parse_board_file(board_file_path)

        return board_info['dP'].sum() / 3600

    def compute_consumption_total_mission(self, mission_folder_path):
        """
        Calculate the total energy consumption of the file

        Parameters
        ----------
        mission_folder_path : str or Path

        Returns
        -------
        Total consumption in the mission
        """
        if not isinstance(mission_folder_path, pathlib.Path):
            mission_folder_path = pathlib.Path(mission_folder_path)

        total_consumption = 0
        for board_file_i in mission_folder_path.glob('**/*.txt'):
            if 'board' in board_file_i.name:
                total_consumption += self.compute_consumption(board_file_i)

        return total_consumption

    @staticmethod
    def read_header(file_path, zip_mode=False):
        """
        Return the parameters of the .wav file's header as a dictionary

        Parameters
        ----------
        file_path: Path or string
            Path to the .wav file to read the header from

        Returns
        -------
        extra_header: dictionary with all the parameters of the configuration provided by RTSys
        """
        HEADER_CONF = {
            'conf': {'format': 'c', 'start': 0, 'end': 4, 'n': 4},
            'conf_size': {'format': 'I', 'start': 4, 'end': 8, 'n': 1},
            'conf_version': {'format': 'I', 'start': 8, 'end': 12, 'n': 1},
            'epoch_time_recording': {'format': 'd', 'start': 16, 'end': 24, 'n': 1},
            'channel': {'format': 'c', 'start': 23, 'end': 24, 'n': 1},
            'sr': {'format': 'f', 'start': 28, 'end': 32, 'n': 1},
            'hydrophone_sensitivity_A': {'format': 'f', 'start': 32, 'end': 36, 'n': 1},
            'hydrophone_sensitivity_B': {'format': 'f', 'start': 36, 'end': 40, 'n': 1},
            'hydrophone_sensitivity_C': {'format': 'f', 'start': 40, 'end': 44, 'n': 1},
            'hydrophone_sensitivity_D': {'format': 'f', 'start': 44, 'end': 48, 'n': 1},
            'hydrophone_amplification_A': {'format': 'f', 'start': 48, 'end': 52, 'n': 1},
            'hydrophone_amplification_B': {'format': 'f', 'start': 52, 'end': 56, 'n': 1},
            'hydrophone_amplification_C': {'format': 'f', 'start': 56, 'end': 60, 'n': 1},
            'hydrophone_amplification_D': {'format': 'f', 'start': 60, 'end': 64, 'n': 1},
            'correction_factor_A': {'format': 'f', 'start': 64, 'end': 68, 'n': 1},
            'correction_factor_B': {'format': 'f', 'start': 68, 'end': 72, 'n': 1},
            'correction_factor_C': {'format': 'f', 'start': 72, 'end': 76, 'n': 1},
            'correction_factor_D': {'format': 'f', 'start': 76, 'end': 80, 'n': 1},
            'serial_number': {'format': 'f', 'start': 80, 'end': 96, 'n': 4},
            'active_channels': {'format': 'c', 'start': 100, 'end': 104, 'n': 4}
        }

        if zip_mode == True:
            path_zip = file_path.split('.zip')[0] + '.zip'
            file_zip = os.path.relpath(file_path, start=path_zip).replace('\\', '/')
            zipFolder = zipfile.ZipFile(path_zip, 'r')
            f = zipFolder.open(file_zip)
        else:
            f = open(file_path, 'rb')
        f.seek(40)
        config_size = struct.unpack('I', f.read(4))[0]
        f.seek(36)
        header_buffer = f.read(config_size)
        f.close()
        extra_header = {}
        for chunk_name, chunk_format in HEADER_CONF.items():
            if (chunk_format['n'] == 1) & (chunk_format['format'] != 'c'):
                extra_header[chunk_name] = struct.unpack(chunk_format['format'],
                                                         header_buffer[chunk_format['start']:chunk_format['end']])[0]
            else:
                extra_header[chunk_name] = bytes.decode(header_buffer[chunk_format['start']:chunk_format['end']])

        return extra_header

    def update_metadata(self, file_path, zip_mode=False):
        """
        Creates a new RTSys object from an already existing one but updating the metadata from the file header.
        The "mode" parameter stays the same.

        Parameters
        ----------
        file_path: str or Path
            path to the wav file recorded with RTSys with a correc header
        zip_mode: bool
            True if file is zipped, otherwise false
        """
        header = self.read_header(file_path, zip_mode)
        name, model, serial_number, sens, ampl = self.meta_from_header(header)
        return RTSys(name=name, model=model, serial_number=serial_number, sensitivity=sens, preamp_gain=ampl, Vpp=5.0,
                     mode=self.mode, calibration_file=self.calibration_file)

    def meta_from_header(self, header):
        sens = header['hydrophone_sensitivity_%s' % self.channel]
        name = 'RTSys'
        model = None
        serial_number = header['serial_number']

        if self.mode == 'lowpower':
            ampl = 20 * np.log10(5 / np.sqrt(2))
        else:
            ampl = 20 * np.log10((1 / (header['hydrophone_amplification_%s' % self.channel] *
                                       header['correction_factor_%s' % self.channel])))
        return name, model, serial_number, sens, ampl

    def one_rtsys_per_channel_from_header(self, file_path, zip_mode=False):
        extra_header = self.read_header(file_path, zip_mode)
        active_channels = []
        for channel_i in np.arange(4):
            if extra_header['active_channels'][channel_i] != '\x00':
                active_channels.append(extra_header['active_channels'][channel_i])

        rtsys_list = []
        for channel in active_channels:
            name, model, serial_number, sens, ampl = self.meta_from_header(extra_header)
            rtsys_list.append(RTSys(name=name, model=model, serial_number=serial_number, sensitivity=sens,
                                    preamp_gain=ampl, Vpp=5.0, channel=channel))

        if len(rtsys_list) == 1:
            return rtsys_list[0]
        else:
            return rtsys_list

    def calibrate(self, file_path, zip_mode=False):
        header = self.read_header(file_path, zip_mode)
        name, model, serial_number, sens, ampl = self.meta_from_header(header)
        self.sensitivity = sens
        self.preamp_gain = ampl
        self.Vpp = 5.0

    def get_freq_cal(self, val='sensitivity', sep=';', freq_col_id=0, val_col_id=1, start_data_id=0):
        super().get_freq_cal(sep=sep, val=val, freq_col_id=freq_col_id, val_col_id=val_col_id,
                             start_data_id=start_data_id)



#!/usr/bin/python
from pyhydrophone.hydrophone import Hydrophone

import os
from datetime import datetime


class AmarG3(Hydrophone):
    def __init__(self, name, model, serial_number, sensitivity, preamp_gain, Vpp, string_format="%Y%m%dT%H%M%S",
                 calibration_file=None, **kwargs):
        """
        Init an instance of AMARG3

        Parameters
        ----------
        name: str
            Name of the acoustic recorder
        model: str or int
            Model of the acoustic recorder
        serial_number : str or int
            Serial number of the acoustic recorder
        sensitivity : float
            Sensitivity of the acoustic recorder in db
        preamp_gain : float
            Gain of the preamplifier in dB
        Vpp : float
            Voltage peak to peak in volts
        string_format : string
            Format of the datetime string present in the filename
        calibration_file : string or Path
            File where the frequency dependent sensitivity values for the calibration are
        """
        super().__init__(name, model, serial_number, sensitivity, preamp_gain, Vpp, string_format, calibration_file,
                         **kwargs)

    def get_name_datetime(self, file_name):
        """
        Get the data and time of recording from the name of the file

        Parameters
        ----------
        file_name : string
            File name (not path) of the file
        """
        name = os.path.split(file_name)[1]
        date_string = name.split('.')[-2][-16:-1]
        date = super().get_name_datetime(date_string)
        return date

    def get_new_name(self, filename, new_date):
        """
        Replace the datetime with the appropriate one

        Parameters
        ----------
        filename : string
            File name (not path) of the file
        new_date : datetime object
            New datetime to be replaced in the filename
        """
        old_date = self.get_name_datetime(filename)
        old_date_name = datetime.strftime(old_date, "%Y%m%dT%H%M%S")
        new_date_name = datetime.strftime(new_date, "%Y%m%dT%H%M%S")
        new_filename = filename.replace(old_date_name, new_date_name)
        
        return new_filename


class AmarG3MEMS(AmarG3):
    def __init__(self, name, model, serial_number, hydroph_sensitivity, preamp_gain, mems_sensitivity, Vpp):
        """
        Add the MEMS specs

        Parameters
        ----------
        name: str
            Name of the acoustic recorder
        model: str or int
            Model of the acoustic recorder
        serial_number : str or int
            Serial number of the acoustic recorder
        hydroph_sensitivity : float
            Sensitivity of the acoustic recorder in db
        preamp_gain : float
            Gain of the preamplifier in dB
        mems_sensitivity : float
            Sensitivity of the accelerometer
        Vpp : float
            Voltage peak to peak in volts
        """
        self.mems_sensitivity = mems_sensitivity
        super().__init__(name, model, serial_number, hydroph_sensitivity, preamp_gain, Vpp)
